# COMPUTER CODE

# A Regularization Corrected Score Method for
# Nonlinear Regression Models with Covariate Error
# David M. Zucker, Malka Gorfine, Yi Li, Mahlet G. Tadesse, and Donna Spiegelman
# Published in Biometrics

# CODE FOR N&S AND SIMEX METHODS FOR DATA EXAMPLE

library("simex",lib.loc="/udd/stzuc/R/sparc-sun-solaris2.8-library/2.8/")
library("nleqslv",lib.loc="/udd/stzuc/R/sparc-sun-solaris2.8-library/2.8/")

#library("nleqslv",lib.loc="/udd/stzuc/R/sparc-sun-solaris2.8-library/2.8/")
options(width=200) #to get long line
options(digits=4) # for printing 4 digits after the decimal point

#clean up working environment from past runs
rm(list = ls())
graphics.off()

#########################################################################################

#CORRECTED SCORE FUNCTION
corscor = function(beta,y,w,z,flg) {

  b0 = beta[1]
  b1 = beta[2]
  sz = sig*z 
  wv = w
  dim(wv) = c(n,1)
  ewv = exp(b0+b1*wv)
  wmat = wv%*%wuns
  ew = ewv%*%wuns   
  
  a = ew * cos(b1*sz)
  b = ew * sin(b1*sz)
  denom = (1+a)^2 + b^2
  f = (a*(1+a)+b^2)/denom
  g = b/denom
  h = wmat*f - sz*g
  
  fmean = rowSums(f)/k
  hmean = rowSums(h)/k
  
  u1sv = y - fmean
  u2sv = w*y - hmean
  
  if (flg == 1) {
    u1s = mean(y-fmean)
    u2s = mean(w*y-hmean)
    return(c(u1s,u2s))
   }
   
  if (flg == 2) {
    c11 = mean(u1sv*u1sv)
    c12 = mean(u1sv*u2sv)
    c22 = mean(u2sv*u2sv)
    return(c(c11,c12,c12,c22))
   }  
      
 }
#END CORRECTED SCORE FUNCTION

###################################################################

# JACOBIAN OF CORRECTED SCORE FUNCTION
csjac = function(beta,y,w,z,flg) {

  jacmat = matrix(0,2,2)
  b0 = beta[1]
  b1 = beta[2]
  sz = sig*z 
  wv = w
  dim(wv) = c(n,1)
  ewv = exp(b0+b1*wv)
  wmat = wv%*%wuns
  ew = ewv%*%wuns   
  
  a = ew * cos(b1*sz)
  b = ew * sin(b1*sz)
  denom = (1+a)^2 + b^2
  f = (a*(1+a)+b^2)/denom
  g = b/denom
  h = wmat*f - sz*g
  
  dadb0 = a
  dadb1 = wmat*a - sz*b
  dbdb0 = b
  dbdb1 = wmat*b + sz*a
  
  dfdb0 = ((1+2*a)*dadb0+2*b*dbdb0)/denom - f*(2*(1+a)*dadb0+2*b*dbdb0)/denom
  dfdb1 = ((1+2*a)*dadb1+2*b*dbdb1)/denom - f*(2*(1+a)*dadb1+2*b*dbdb1)/denom
  dgdb1 = dbdb1/denom - g*(2*(1+a)*dadb1+2*b*dbdb1)/denom
  dhdb1 = wmat*dfdb1 - sz*dgdb1
  
  jacmat[1,1] = -mean(rowSums(dfdb0)/k)
  jacmat[1,2] = -mean(rowSums(dfdb1)/k)
  jacmat[2,1] = jacmat[1,2]
  jacmat[2,2] = -mean(rowSums(dhdb1)/k)
  return(jacmat)

 } 
#END JACOBIAN OF CORRECTED SCORE FUNCTION 

###################################################################

# SANDWICH VARIANCE FOR N&S ESTIMATOR

sandvar = function(betest,y,w,z) {
  jacmat = csjac(betest,y,w,z,1)
  kap = kappa(jacmat, exact=TRUE)
  jacinv = solve(jacmat)
  filling = corscor(betest,y,w,z,2)
  dim(filling) = c(2,2)
  sandvar = jacinv %*% filling %*% jacinv / n
 } 

###################################################################


#MAIN PROGRAM

# SETTINGS
k = 100
wuns = rep(1,k)

# RESULTS FROM VALIDATION DATA
xwint = 1.22701
xwslp = 0.56533
vare = 0.81812/(xwslp^2)
sig = sqrt(vare)

# MAIN DATA
data = read.table("main1.dat")
y = data[,1]
w = data[,2]
n = length(w)
w = (w-xwint)/xwslp

# NAIVE ESTIMATE
naive = glm(y~w,family=binomial,x=TRUE)
betnai = c(naive$coef[1],naive$coef[2])
print(naive)

#SIMEX ESTIMATES
simex1 <- simex(model = naive , SIMEXvariable = "w", B=200, measurement.error = sig,
fitting.method = "quadratic", jackknife.estimation = "quadratic" ,asymptotic = FALSE)
simex2 <- refit(simex1,"linear", jackknife.estimation = "linear" ,asymptotic=FALSE)
simex3 <- try(refit(simex1,"nonlinear", jackknife.estimation = "nonlinear",asymptotic=FALSE),TRUE)
print(simex1$coef)
print(simex2$coef)
print(simex3$coef)
print(simex1$variance.jackknife[1,1]) #intercept
print(simex1$variance.jackknife[2,2]) #slope
print(simex2$variance.jackknife[1,1])
print(simex2$variance.jackknife[2,2])
print(simex3$variance.jackknife[1,1])
print(simex3$variance.jackknife[2,2])
print(sqrt(simex1$variance.jackknife[1,1])) #intercept
print(sqrt(simex1$variance.jackknife[2,2])) #slope
print(sqrt(simex2$variance.jackknife[1,1]))
print(sqrt(simex2$variance.jackknife[2,2]))

# N&S ESTIMATE
z = rnorm(n,0,1)
for (kk in 2:k) {
  zz = rnorm(n,0,1)
  z = cbind(z,zz)
 }
beta = betnai
cntl = list(1e-6,1e-6,1e-3,100)
names(cntl) = c('xtol','ftol','btol','maxit')
funest = nleqslv(beta,corscor,jac=csjac,y=y,w=w,z=z,flg=1,control=cntl)
betest = funest$x
print(betest)
theocov = sandvar(betest,y,w,z)
sds = sqrt(diag(theocov))
print(sds)


