# COMPUTER CODE

# A Regularization Corrected Score Method for
# Nonlinear Regression Models with Covariate Error
# David M. Zucker, Malka Gorfine, Yi Li, Mahlet G. Tadesse, and Donna Spiegelman
# Published in Biometrics

# PROGRAM TO RUN SIMULATIONS OF N&S METHOD (UNIVARIATE)

library(nleqslv)
#library("nleqslv",lib.loc="/udd/stzuc/R/sparc-sun-solaris2.8-library/2.8/")
options(width=200) #to get long line

#clean up working environment from past runs
rm(list = ls())
graphics.off()

# n = main study sample size
# nex = external validation study sample size
# trubet = true beta vector
# samples = number of simulation replications

# truscen = "true" scenario used to generate the data
#   1 = normal-normal homoscedastic
#   2 = normal-normal heteroscedastic
#   3 = normal-normal heteroscedastic with differential error
#   4 = normal - double exponential
#   5 = normal-modchi2
#   6 = modchi2-modchi2
#   7 = normal - skewed normal    
#   8 = modchi2 - skewed normal
#   9 = skewed normal - skewed normal
#  10 = skewed normal - modchi2

# ea, eb, ec = meas error parameters
# (when eb and/or ec are not relevant, set them to NULL)

# extrfac = cutoff for designating result as nonconvergence
# xtol, ftol = tolerances in the nonlinear systems solver for solving the corrected score equations

#INPUTS
n = 200
nex = 70
truscen = 1
trubet = c(1,1)
extrfac = 5
ea = 1
eb = NULL
ec = NULL
k = 100
samples = 1000
xtol = 1e-6
ftol = 5e-3
btol = 1e-3
maxit = 200

#PARAMETERS FOR MODCHI2 AND SN DISTNS
mdchmean = 0.8497415
mdchsd = 1.037242
snlambda = 50
sndel = sqrt(1/(snlambda^(-2)+1))
snlam = sndel/sqrt(1-sndel^2)
snmean = sqrt(2/pi)*sndel
snsd = sqrt(1-snmean^2)

#VECTOR OF K ONES     
wuns = rep(1,k)


#########################################################################################

#ROUTINE TO GENERATE DATA FOR THE SELECTED SCENARIO
dataset = function (nobs) {

  if (truscen == 1) {
    #NORMAL-NORMAL HOMOSCEDASTIC
    x = rnorm(nobs,0,1)
    oddrat = exp(trubet[1] + trubet[2]*x)
    pr = oddrat/(1+oddrat)
    y = rbinom(nobs,1,pr)
    condsig = sqrt(ea)
    e = rnorm(nobs,0,sd=condsig)
    w = x + e
   }

  if (truscen == 2) {
    #NORMAL-NORMAL HETEROSCEDASTIC
    x = rnorm(nobs,0,1)
    oddrat = exp(trubet[1] + trubet[2]*x)
    pr = oddrat/(1+oddrat)
    y = rbinom(nobs,1,pr)
    condsig = sqrt(ea+eb*abs(x))
    e = rnorm(nobs,0,sd=condsig)
    w = x + e
   }

  if (truscen == 3) {
    #NORMAL-NORMAL DIFFL ERROR
    x = rnorm(nobs,0,1)
    oddrat = exp(trubet[1] + trubet[2]*x)
    pr = oddrat/(1+oddrat)
    y = rbinom(nobs,1,pr)
    condsig = sqrt(ea+eb*abs(x)+ec*y)
    e = rnorm(nobs,0,sd=condsig)
    w = x + e
   }

  if (truscen == 4) {
    #LAPLACE (DBLE EXPL) ERROR
    x = rnorm(nobs,0,1)
    oddrat = exp(trubet[1] + trubet[2]*x)
    pr = oddrat/(1+oddrat)
    y = rbinom(nobs,1,pr)
    sig = sqrt(ea)
    exprte = sqrt(2)/sig
    eabs = rexp(nobs,exprte)
    esign = 2*rbinom(nobs,1,0.5)-1
    e = esign*eabs
    w = x + e
   }

  if (truscen == 5) {
    #NORMAL-MODCHI
    sig = sqrt(ea)
    x = rnorm(nobs,0,1)
    oddrat = exp(trubet[1] + trubet[2]*x)
    pr = oddrat/(1+oddrat)
    y = rbinom(nobs,1,pr)
    urv = 0.9746527*runif(nobs)
    e = qchisq(urv,1)
    e = (e - mdchmean)/mdchsd
    e = sig*e
    w = x + e
  }

   if (truscen == 6) {
     #MODCHI-MODCHI
     sig = sqrt(ea)
     urv1 = 0.9746527*runif(nobs)
     x = qchisq(urv1,1)
     x = (x - mdchmean)/mdchsd
     oddrat = exp(trubet[1] + trubet[2]*x)
     pr = oddrat/(1+oddrat)
     y = rbinom(nobs,1,pr)    
     urv2 = 0.9746527*runif(nobs)
     e = qchisq(urv2,1)
     e = (e - mdchmean)/mdchsd
     e = sig*e
     w = x + e
    } 
   
   if (truscen == 7) {
     #NORMAL-SN
     sig = sqrt(ea)
     x = rnorm(nobs,0,1)
     oddrat = exp(trubet[1] + trubet[2]*x)
     pr = oddrat/(1+oddrat)
     y = rbinom(nobs,1,pr)
     sntmp = rmvnorm(nobs,mean=c(0,0),sigma=matrix(c(1,sndel,sndel,1),2))
     e = ifelse(sntmp[,1]>0,sntmp[,2],-sntmp[,2])
     e = (e - snmean)/snsd
     e = sig*e
     w = x + e
    }
   
   if (truscen == 8) {
     #MODCHI-SN
     sig = sqrt(ea)
     urv = 0.9746527*runif(nobs)
     x = qchisq(urv,1)
     x = (x - mdchmean)/mdchsd
     oddrat = exp(trubet[1] + trubet[2]*x)
     pr = oddrat/(1+oddrat)
     y = rbinom(nobs,1,pr)
     sntmp = rmvnorm(nobs,mean=c(0,0),sigma=matrix(c(1,sndel,sndel,1),2))
     e = ifelse(sntmp[,1]>0,sntmp[,2],-sntmp[,2])
     e = (e - snmean)/snsd
     e = sig*e
     w = x + e
    }
     
   if (truscen == 9) {
     #SN-SN
     sig = sqrt(ea)
     sntmp = rmvnorm(nobs,mean=c(0,0),sigma=matrix(c(1,sndel,sndel,1),2))
     x = ifelse(sntmp[,1]>0,sntmp[,2],-sntmp[,2])
     x = (x - snmean)/snsd
     oddrat = exp(trubet[1] + trubet[2]*x)
     pr = oddrat/(1+oddrat)
     y = rbinom(nobs,1,pr)
     sntmp = rmvnorm(nobs,mean=c(0,0),sigma=matrix(c(1,sndel,sndel,1),2))
     e = ifelse(sntmp[,1]>0,sntmp[,2],-sntmp[,2])
     e = (e - snmean)/snsd
     e = sig*e
     w = x + e
    }    
     
   if (truscen == 10) {
     #SN-MODCHI
     sig = sqrt(ea)
     sntmp = rmvnorm(nobs,mean=c(0,0),sigma=matrix(c(1,sndel,sndel,1),2))
     x = ifelse(sntmp[,1]>0,sntmp[,2],-sntmp[,2])
     x = (x - snmean)/snsd
     oddrat = exp(trubet[1] + trubet[2]*x)
     pr = oddrat/(1+oddrat)
     y = rbinom(nobs,1,pr)
     urv = 0.9746527*runif(nobs)
     e = qchisq(urv,1)
     e = (e - mdchmean)/mdchsd
     e = sig*e
     w = x + e
    }       
     
   dim(y) = c(nobs,1)
   dim(w) = c(nobs,1)
   dim(x) = c(nobs,1)
   dataset = cbind(y,w,x)
  
 } 

#########################################################################################

# CORRECTED SCORE FUNCTION
corscor = function(beta,y,w,z,flg) {

  b0 = beta[1]
  b1 = beta[2]
  sz = sig*z 
  wv = w
  dim(wv) = c(n,1)
  ewv = exp(b0+b1*wv)
  wmat = wv%*%wuns
  ew = ewv%*%wuns   
  
  a = ew * cos(b1*sz)
  b = ew * sin(b1*sz)
  denom = (1+a)^2 + b^2
  f = (a*(1+a)+b^2)/denom
  g = b/denom
  h = wmat*f - sz*g
  
  fmean = rowSums(f)/k
  hmean = rowSums(h)/k
  
  u1sv = y - fmean
  u2sv = w*y - hmean
  
  if (flg == 1) {
    u1s = mean(y-fmean)
    u2s = mean(w*y-hmean)
    return(c(u1s,u2s))
   }
   
  if (flg == 2) {
    c11 = mean(u1sv*u1sv)
    c12 = mean(u1sv*u2sv)
    c22 = mean(u2sv*u2sv)
    return(c(c11,c12,c12,c22))
   }  
      
 }
#END CORRECTED SCORE FUNCTION

# JACOBIAN OF CORRECTED SCORE FUNCTION
csjac = function(beta,y,w,z,flg) {

  jacmat = matrix(0,2,2)
  b0 = beta[1]
  b1 = beta[2]
  sz = sig*z 
  wv = w
  dim(wv) = c(n,1)
  ewv = exp(b0+b1*wv)
  wmat = wv%*%wuns
  ew = ewv%*%wuns   
  
  a = ew * cos(b1*sz)
  b = ew * sin(b1*sz)
  denom = (1+a)^2 + b^2
  f = (a*(1+a)+b^2)/denom
  g = b/denom
  h = wmat*f - sz*g
  
  dadb0 = a
  dadb1 = wmat*a - sz*b
  dbdb0 = b
  dbdb1 = wmat*b + sz*a
  
  dfdb0 = ((1+2*a)*dadb0+2*b*dbdb0)/denom - f*(2*(1+a)*dadb0+2*b*dbdb0)/denom
  dfdb1 = ((1+2*a)*dadb1+2*b*dbdb1)/denom - f*(2*(1+a)*dadb1+2*b*dbdb1)/denom
  dgdb1 = dbdb1/denom - g*(2*(1+a)*dadb1+2*b*dbdb1)/denom
  dhdb1 = wmat*dfdb1 - sz*dgdb1
  
  jacmat[1,1] = -mean(rowSums(dfdb0)/k)
  jacmat[1,2] = -mean(rowSums(dfdb1)/k)
  jacmat[2,1] = jacmat[1,2]
  jacmat[2,2] = -mean(rowSums(dhdb1)/k)
  return(jacmat)

 } 
#END JACOBIAN OF CORRECTED SCORE FUNCTION 

#DERIVATIVE OF CORRECTED SCORE FUNCTION W.R.T. SIGMA
cdsig = function(beta,y,w,z) {

  b0 = beta[1]
  b1 = beta[2]
  sz = sig*z 
  wv = w
  dim(wv) = c(n,1)
  ewv = exp(b0+b1*wv)
  wmat = wv%*%wuns
  ew = ewv%*%wuns   
  
  a = ew * cos(b1*sz)
  b = ew * sin(b1*sz)
  denom = (1+a)^2 + b^2
  f = (a*(1+a)+b^2)/denom
  g = b/denom
  h = wmat*f - sz*g
  
  d_a_sig = -b1*z*b
  d_b_sig = b1*z*a
  
  d_f_sig = ((1+2*a)*d_a_sig+2*b*d_b_sig)/denom - f*(2*(1+a)*d_a_sig+2*b*d_b_sig)/denom
  d_g_sig = d_b_sig/denom - g*(2*(1+a)*d_a_sig+2*b*d_b_sig)/denom 
  d_h_sig = wmat*d_f_sig - sz*d_g_sig - z*g
  
  cds1 = -mean(rowSums(d_f_sig)/k)
  cds2 = -mean(rowSums(d_h_sig)/k)
  return(c(cds1,cds2))

 }
# END DERIVATIVE OF CORRECTED SCORE FUNCTION W.R.T. SIGMA 

  
#MAIN PROGRAM 

#PRINT OUT INPUTS
inputs1 = cbind(n,nex,samples,truscen,ea,eb,ec)
print(inputs1)
inputs2 = cbind(xtol,ftol,btol,maxit)
print(inputs2)
print("trubet")
print(trubet)

#INITIALIZE RANDOM NUMBER GENERATOR
seed0 = seq(1,samples)+11817

#SET UP RESULTS MATRICES
estimates = array(rep(NA,4*samples),dim=c(samples,2,2))
sds = array(rep(NA,4*samples),dim=c(samples,2,2))
cicov = array(rep(NA,4*samples),dim=c(samples,2,2))
estmat = matrix(NA,2,2)
emmat = matrix(NA,2,2)
sdmat = matrix(NA,2,2)
sdmmat = matrix(NA,2,2)
empsdmat = matrix(NA,2,2)
iqrmat = matrix(NA,2,2)
cimat = matrix(NA,2,2)
gudnum = c(0,0)

#RUN SIMULATIONS
for (i in 1:samples) {

#GENERATE DATA
set.seed(seed0[i])
dtaset = dataset(n)
y = dtaset[,1]
w = dtaset[,2]

#GENERATE VALIDATION DATA AND ESTIMATE OF SIGMA
#sig=1
#varsig=0 
dtaset0 = dataset(nex)
wvld = dtaset0[,2]
xvld = dtaset0[,3]
ervld = wvld-xvld
varbas = sum(ervld^2)/nex
vvb = 2*(varbas^2)/nex
sig = sqrt(varbas)
varsig = vvb/(4*varbas)
 
# GENERATE Z'S
z = rnorm(n,0,1)
for (kk in 2:k) {
  zz = rnorm(n,0,1)
  z = cbind(z,zz)
 }

# NAIVE ESTIMATE
naive = glm(y~w,family=binomial)
betnai = c(naive$coef[1],naive$coef[2])
covnai = summary(naive)$cov.unscaled
if (abs(betnai[2]) <= 5) {
gudnum[1] = gudnum[1] + 1
estimates[i,1,1] = betnai[1]
estimates[i,2,1] = betnai[2]
sds[i,1,1] = sqrt(covnai[1,1])
sds[i,2,1] = sqrt(covnai[2,2])
for (rr in 1:2) {
  diff = abs(estimates[i,rr,1]-trubet[rr]) - 1.96*sds[i,rr,1]
  cicov[i,rr,1] = (diff <= 0)
 }
} 

# COMPUTE CORRECTED SCORE ESTIMATOR
beta = betnai 
cntl = list(xtol,ftol,btol,maxit)
names(cntl) = c('xtol','ftol','btol','maxit')
funest = nleqslv(beta,corscor,jac=csjac,y=y,w=w,z=z,flg=1,control=cntl)
bet0 = funest$x[1]
bet1 = funest$x[2]
fvec = funest$fvec
termcd = funest$termcd
fvamax = max(abs(fvec))
good = (fvamax <= ftol)

# WEED OUT BAD/EXTREME ESTIMATES
if (good & (abs(bet1) <= extrfac)) {
#if (abs(bet1) <= extrfac) {
#if ((termcd == 1) & (abs(bet1) <= extrfac)) {

gudnum[2]= gudnum[2] + 1

#store estimates
estimates[i,1,2] = bet0
estimates[i,2,2] = bet1

#compute theoretical variances and ci coverage indicators
betest = c(bet0,bet1)
jacmat = csjac(betest,y,w,z,1)
kap = kappa(jacmat, exact=TRUE)
jacinv = solve(jacmat)
filling = corscor(betest,y,w,z,2)
dim(filling) = c(2,2)
theocov = jacinv %*% filling %*% jacinv / n
cdsvec = cdsig(betest,y,w,z)
#print(cdsvec)
#sig = sig + 0.0001
#scor1 = corscor(betest,y,w,z,1)
#sig = sig - 0.0002
#scor2 = corscor(betest,y,w,z,1)
#cdsv = (scor1-scor2)/0.0002
#print(cdsv)
dim(cdsvec) = c(2,1)
covtrm = varsig * (cdsvec %*% t(cdsvec))
theocov = theocov + covtrm
sds[i,1,2] = sqrt(theocov[1,1])
sds[i,2,2] = sqrt(theocov[2,2])
for (rr in 1:2) {
  diff = abs(estimates[i,rr,2]-trubet[rr]) - 1.96*sds[i,rr,2]
  cicov[i,rr,2] = (diff <= 0)
 }
}

#dim(jacmat) = c(1,4)
#dim(filling) = c(1,4)
#print(c(bet0,bet1,jacmat,kap,filling,sds[i,1,2],sds[i,2,2]))

}

#SUMMARY STATISTICS
pctgud = gudnum/samples
for (imeth in 1:2) {
for (rr in 1:2) {
  estvec = estimates[,rr,imeth]
  sdvec = sds[,rr,imeth]
  civec = cicov[,rr,imeth]
  #print("Statistics on Estimates")
  #print(stem(estvec))
  #print(summary(estvec,na.rm=TRUE))
  #print("Statistics on Theoretical SD's")
  #print(stem(sdvec))
  #print(summary(sdvec,na.rm=TRUE))  
  estmat[rr,imeth] = mean(estvec,na.rm=TRUE)
  emmat[rr,imeth] = median(estvec,na.rm=TRUE)
  sdmat[rr,imeth] = mean(sdvec,na.rm=TRUE)
  sdmmat[rr,imeth] = median(sdvec,na.rm=TRUE)
  empsdmat[rr,imeth] = sd(estvec,na.rm=TRUE)
  sumstat = summary(estvec,na.rm=TRUE)
  iqrmat[rr,imeth] = (sumstat[5]-sumstat[2])/1.349
  cimat[rr,imeth] = mean(civec,na.rm=TRUE)
 }
 }

gud = pctgud
ee = data.frame(estmat)
eem = data.frame(emmat)
ss1 = data.frame(sdmat)
ss1a = data.frame(sdmmat)
ss2 = data.frame(empsdmat)
iqr = data.frame(iqrmat)
ci = data.frame(cimat)
colname = c("naive","mc-ce")
names(gud) = colname
names(ee) = colname
names(eem) = colname
names(ss1) = colname
names(ss1a) = colname
names(ss2) = colname
names(iqr) = colname
names(ci) = colname
print("Percent of Samples with Successful Estimation")
print(gud)
print("Mean of Estimates")
print(ee)
print("Median of Estimates")
print(eem)
print("Mean Estimated SD")
print(ss1)
print("Median Estimated SD")
print(ss1a)
print("Empirical SD of Estimates")
print(ss2)
print("IQR/1.349")
print(iqr)   
print("Confidence Interval Coverage")
print(ci) 
