#RECS PROGRAM
#source("recs-example.r")

#########################################################################################
#THIS CODE MAKES USE OF 5 EXTERNAL FORTRAN ROUTINES, AS FOLLOWS
#  setup.f:     sets up various matrices needed for the computations
#  delcmp.f:    computes the average delta vector at the naive beta
#  scoresoln.f: finds solution to corrected score equation
#                 (includes routines to compute the corrected score and its jacobian, 
#                  along with an existing nonlinear system solver routine)
#  scorefn.f:   computes the corrected score function
#                 (returns either individual-subject results or summary results)
#  scorejac.f:  computes the jacobian of the score function (summed over subjects)
#########################################################################################     

#SETUPS

#clean up working environment from past runs
rm(list = ls())
graphics.off()

#libraries, fortran codes, and options
pcrun = 0
options(width=200)
library(statmod)
if (pcrun == 1) {
  library(orthopolynom)
  dyn.load("setup.dll")
  dyn.load("scoresoln.dll")
  dyn.load("scorefn.dll")
  dyn.load("scorejac.dll")
  dyn.load("delcmp.dll")  
 } else { 
  library("polynom",lib.loc="/udd/stzuc/R/sparc-sun-solaris2.8-library/2.8/")
  library("orthopolynom",lib.loc="/udd/stzuc/R/sparc-sun-solaris2.8-library/2.8/")
  dyn.load("setup.so")
  dyn.load("scoresoln.so")
  dyn.load("scorefn.so")
  dyn.load("scorejac.so")
  dyn.load("delcmp.so")  
 }

# n = main study sample size
# nex = external validation study sample size
# p = number of error-prone covariates
# pz = number of error-free covariates

# ialmin = flag for alpha selection
#   ialmin=0 means use the default value
#   ialmin=1 means use the GCV selection method

# truscen = "true" scenario used to generate the data
#   1 = normal-normal homoscedastic
#   2 = normal-normal heteroscedastic
#   3 = normal-normal heteroscedastic with differential error
#   4 = normal - double exponential
#   5 = normal-modchi2
#   6 = modchi2-modchi2
#   7 = normal - skewed normal

# iscen = "assumed" scenario used in the analysis
#  (same codes as above, except code 7 is not used)

# ixiest = flag for estimation of meas error parameters
#   ixiest=0 means take the input meas error parameters as known values
#   ixiest=1 means estimate the meas error parameters from external validation data

# ea, eb, ec = meas error parameters
# (when eb and/or ec are not relevant, set them to NULL)

# extrfac = cutoff for designating result as nonconvergence

# kk1 = number of quad nodes in quadrature for A-operator integral
# kk2 = number of quad nodes in quadrature for norm used in objective function
# dfr = number of basis functions (per dimension) in basis expansion
# alfdef = default alpha value
# alfstep = number of steps in scheme for stepping up alpha value when there are convergence problems
# xtol, ftol = tolerances in the nonlinear systems solver for solving the corrected score equations

# BASIC INPUTS
p = 1
pz = 0
ialmin = 1
ixiest = 0
extrfac = Inf
iscen = 1
nex = 149

#BASIC SETTINGS
kk1 = 20
kk2 = 20
dfr = 6
alfdef = 0.01
alfstp = 50
xtol = 1e-6
ftol = 1e-4

#PARAMETERS FOR MODCHI2 AND SN DISTNS
mdchmean = 0.8497415
mdchsd = 1.037242
snlambda = 50
sndel = sqrt(1/(snlambda^(-2)+1))
snlam = sndel/sqrt(1-sndel^2)
snmean = sqrt(2/pi)*sndel
snsd = sqrt(1-snmean^2)

if (iscen == 4) {
  odd = kk1 %% 2
  if (odd) {
    kk1 = kk1 + 1
    print("kk1 increased by 1 to make it an even number")
    print(" ")
   }
 }  

#########################################################################################

#DIGITSBASE RECONSTITUTED
digi = function(valvec,base,ndig) {
  lval = length(valvec)
  digi = matrix(0,lval,ndig)
  for (iv in 1:lval) {
    val = valvec[iv]-1
    maxpwr = 0
    if (val>0) {maxpwr = floor(log(val)/log(base))}
    vchk = base^(maxpwr+1)
    if (vchk <= val) {maxpwr = maxpwr + 1}
    curval = val
    for (ipx in 0:maxpwr) {
      cpwr = maxpwr - ipx
      cnum = base^cpwr
      cdig = curval %/% cnum
      digi[iv,ndig-cpwr] = cdig
      curval = curval %% cnum
    }
   }
  digi = digi + 1
  return(digi)
 }

#########################################################################################

#ORTHOGONAL POLYNOMIAL BASIS FUNCTION SETUP
pbas = function(dfr) { 
  pb = matrix(0,dfr,dfr)
  dfr1 = dfr-1
  orthplist = hermite.he.polynomials(dfr1,normalized=FALSE)
  coeflst = polynomial.coefficients(orthplist)
  coefvec = unlist(coeflst)
  count = 0
  for (k in 1:dfr) {
  for (l in 1:k) {
    count = count + 1
    pb[k,l] = coefvec[count]
   }
   }
  fac = sqrt(1/factorial(c(0:dfr1)))
  pb = diag(fac) %*% pb 
  pbas = pb 
 }

#########################################################################################

#ROUTINE TO SET UP RELEVANT MATRICES
setup = function(xivec) {
  
  #PREPARATIONS
  storage.mode(iscen) = "integer"
  storage.mode(n) = "integer"
  storage.mode(p) = "integer"
  storage.mode(npxi) = "integer"
  storage.mode(dfr) = "integer"
  storage.mode(pdfr) = "integer"
  storage.mode(kk1) = "integer"
  storage.mode(pkk1) = "integer"
  storage.mode(kk2) = "integer"
  storage.mode(pkk2) = "integer"
  storage.mode(y) = "double"
  storage.mode(w) = "double"
  storage.mode(xivec) = "double"
  storage.mode(digmatpdfr) = "integer"
  storage.mode(digmatpkk1) = "integer"
  storage.mode(digmatpkk2) = "integer"
  storage.mode(nodes1) = "double"
  storage.mode(qv1) = "double"
  storage.mode(nodes2) = "double"
  storage.mode(wscal) = "double"
  storage.mode(coefmat) = "double"
  psiw = matrix(0,n,pdfr)  
  phi = array(rep(0,n*pkk2*pdfr),dim=c(n,pkk2,pdfr))
  cdwt = array(rep(0,n*pkk1*pkk2),dim=c(n,pkk2,pkk1))
  wwvec = array(rep(0,n*pkk1*pkk2*p),dim=c(n,pkk2,pkk1,p))
  storage.mode(psiw) = "double"
  storage.mode(phi) = "double"
  storage.mode(cdwt) = "double"
  storage.mode(wwvec) = "double"
  
  #COMPUTATION OF PSIW, PHI, CDWT, AND WWVEC
  forres = .Fortran("setup",iscen,n,p,npxi,dfr,pdfr,kk1,pkk1,kk2,pkk2,y,w,xivec,digmatpdfr,digmatpkk1,
    digmatpkk2,nodes1,qv1,nodes2,wscal,coefmat,psiw=psiw,phi=phi,cdwt=cdwt,wwvec=wwvec)     
  psiw = forres$psiw  
  phi = forres$phi
  cdwt = forres$cdwt
  wwvec = forres$wwvec
  
  #PHITIL MATRIX
  if (phiscen==0) {
    phimat = phi[1,,]
    phitil = qhmat %*% phimat
    sss = svd(phitil)
    uuu1 = sss$u
    vvv1 = sss$v
    singval1 = sss$d
    lambda1 = singval1^2
    for (io in 1:n) {
      uuu[io,,] = uuu1
      vvv[io,,] = vvv1
      singval[io,] = singval1
      lambda[io,] = lambda1
     }
   }  
  if (phiscen==1) {
    for (io in 1:n) {
      phimat = phi[io,,]
      phitil = qhmat %*% phimat
      sss = svd(phitil)
      uuu1 = sss$u
      vvv1 = sss$v
      singval1 = sss$d
      lambda1 = singval1^2
      uuu[io,,] = uuu1
      vvv[io,,] = vvv1
      singval[io,] = singval1
      lambda[io,] = lambda1
     }
   }    
  
  results = list(psiw,phi,cdwt,wwvec,uuu,vvv,singval,lambda)
  names(results) = c('psiw','phi','cdwt','wwvec','uuu','vvv','singval','lambda')
  return(results)

 }

#########################################################################################
 
#ROUTINE TO COMPUTE DELTA
delcmp = function(b,cdwt,wwvec) {
  storage.mode(b) = "double"
  storage.mode(n) = "integer"
  storage.mode(p) = "integer"
  storage.mode(pz) = "integer"
  storage.mode(r) = "integer"
  storage.mode(kk2) = "integer"
  storage.mode(pkk1) = "integer"
  storage.mode(pkk2) = "integer"
  storage.mode(y) = "double"
  storage.mode(z) = "double"
  storage.mode(digmatpkk2) = "integer"
  storage.mode(nodes2) = "double"
  storage.mode(cdwt) = "double"
  storage.mode(wwvec) = "double"
  del = array(rep(0,n*r*pkk2),dim=c(n,r,pkk2))
  storage.mode(del) = "double" 
  delta = .Fortran("delcmp",b,n,p,pz,r,kk2,pkk1,pkk2,y,z,digmatpkk2,
    nodes2,cdwt,wwvec,del=del)
  delta = delta$del
  return(delta) 
 }    

#########################################################################################

#ROUTINE TO COMPUTE GCV CRITERION
gcv = function(alf,lam,mu,delnrm2) {
  gcvmat = matrix(0,n,r)
  for (i in 1:n) {
  for (ir in 1:r) {
    lam1 = lam[i,]
    mu1 = mu[i,ir,]
    filfac = lam1/(lam1+alf)   
    brk0 = filfac^2 - 2*filfac
    e = sum(mu1*brk0) + delnrm2[i,ir]
    trais = pkk2-sum(filfac)
    gcvmat[i,ir] = e/(trais^2)
   }
   }
  gcv = mean(gcvmat)
 }
 
######################################################################################### 

#ROUTINE TO FIND SOLUTION TO CORRECTED SCORE EQUATION
scoresoln = function(b0,cmi,psiw,cdwt,wwvec) { 
  storage.mode(b0) = "double"
  storage.mode(n) = "integer"
  storage.mode(p) = "integer"
  storage.mode(pz) = "integer"
  storage.mode(r) = "integer"
  storage.mode(pdfr) = "integer"
  storage.mode(kk2) = "integer"
  storage.mode(pkk1) = "integer"
  storage.mode(pkk2) = "integer"
  storage.mode(y) = "double"
  storage.mode(w) = "double"
  storage.mode(z) = "double"
  storage.mode(cmi) = "double"
  storage.mode(psiw) = "double"
  storage.mode(digmatpkk2) = "integer"
  storage.mode(nodes2) = "double"
  storage.mode(cdwt) = "double"
  storage.mode(wwvec) = "double"
  storage.mode(xtol) = "double"
  storage.mode(ftol) = "double"  
  b = as.vector(rep(0,r))
  storage.mode(b) = "double"
  igood = 1
  storage.mode(igood) = "integer"
  soln = .Fortran("scoresoln",b0,n,p,pz,r,pdfr,kk2,pkk1,pkk2,y,w,z,cmi,
    psiw,digmatpkk2,nodes2,cdwt,wwvec,xtol,ftol,b=b,igood=igood)
  b = soln$b
  igood = soln$igood
  result = list(b,igood)
  names(result) = c('b','igood')
  return(result)
 }  

#########################################################################################

#SCORE FUNCTION
score = function(b,cmi,psiw,cdwt,wwvec,flag) {
  storage.mode(b) = "double" 
  storage.mode(n) = "integer"
  storage.mode(p) = "integer"
  storage.mode(pz) = "integer"
  storage.mode(r) = "integer"
  storage.mode(pdfr) = "integer"
  storage.mode(kk2) = "integer"
  storage.mode(pkk1) = "integer"
  storage.mode(pkk2) = "integer"
  storage.mode(y) = "double"
  storage.mode(w) = "double"
  storage.mode(z) = "double"
  storage.mode(cmi) = "double"
  storage.mode(psiw) = "double"
  storage.mode(digmatpkk2) = "integer"
  storage.mode(nodes2) = "double"  
  storage.mode(cdwt) = "double"
  storage.mode(wwvec) = "double"
  umat = matrix(0,n,r)
  storage.mode(umat) = "double"
  um = .Fortran("scorefn",b,n,p,pz,r,pdfr,kk2,pkk1,pkk2,y,w,z,
    cmi,psiw,digmatpkk2,nodes2,cdwt,wwvec,umat=umat)
  umat = um$umat  
  if (flag==1)  result = colSums(umat)
  if (flag==2)  result = t(umat)%*%umat
  if (flag==3)  result = umat
  return(result)
 }  
 
######################################################################################### 

#JACOBIAN OF SCORE FUNCTION (SUMMED OVER SUBJECTS)
scorejac = function(b,cmi,psiw,cdwt,wwvec) {
  storage.mode(b) = "double" 
  storage.mode(n) = "integer"
  storage.mode(p) = "integer"
  storage.mode(pz) = "integer"
  storage.mode(r) = "integer"
  storage.mode(pdfr) = "integer"
  storage.mode(kk2) = "integer"
  storage.mode(pkk1) = "integer"
  storage.mode(pkk2) = "integer"
  storage.mode(y) = "double"
  storage.mode(w) = "double"
  storage.mode(z) = "double"
  storage.mode(cmi) = "double"
  storage.mode(psiw) = "double"
  storage.mode(digmatpkk2) = "integer"
  storage.mode(nodes2) = "double"
  storage.mode(cdwt) = "double"
  storage.mode(wwvec) = "double" 
  scjac = matrix(0,r,r)
  storage.mode(scjac) = "double"
  sjres = .Fortran("scorejac",b,n,p,pz,r,pdfr,kk2,pkk1,pkk2,
    y,w,z,cmi,psiw,digmatpkk2,nodes2,cdwt,wwvec,scjac=scjac)
  scjac = sjres$scjac
  return(scjac)
 }  
 
######################################################################################### 

#FUNCTION TO COMPUTE RECS ESTIMATOR
recs  =  function(y,w,z) {

#SOME SETUPS
cmi = array(rep(0,n*pdfr*pkk2),dim=c(n,pdfr,pkk2))
ubard = matrix(0,r,npxi)
estfin = matrix(NA,2,r)
sefin = matrix(NA,2,r)
cicov = matrix(NA,2,r)
didl = 0.0001
  
#NAIVE ESTIMATOR
if (pz>0) {
  err = glm(y~w+z,family=binomial(link='logit'))
 } else {  
  err = glm(y~w,family=binomial(link='logit'))
 }  
estval = err$coef
estcov = summary(err)$cov.unscaled
naiest = trnsmat%*%estval
naicov = trnsmat%*%estcov%*%t(trnsmat)
estfin[1,] = naiest
sefin[1,] = sqrt(diag(naicov))

#SET UP PHIW, PHI, CDWT, WWVEC, UUU, VVV, SINGVAL, AND LAMBDA MATRICES
settings = setup(xi)
psiw = settings$psiw
phi = settings$phi
cdwt = settings$cdwt
wwvec = settings$wwvec
uuu = settings$uuu
vvv = settings$vvv
singval = settings$singval
lambda = settings$lambda
 
#IF DESIRED, SELECT ALPHA BY GCV METHOD
if (ialmin == 0) {alfmin = alfdef}
if (ialmin == 1) {
  mu = array(rep(0,n*r*pdfr),dim=c(n,r,pdfr))
  delnrm2 = matrix(0,n,r)
  del = delcmp(estval,cdwt,wwvec)
  for (i in 1:n) {
    uumat = uuu[i,,]
    for (ir in 1:r) {
      delvec = del[i,ir,]
      deltil = sqqv2new*delvec
      delnrm2[i,ir] = sum(deltil^2)
      rotdeltil = t(uumat) %*% deltil
      mu[i,ir,] = rotdeltil^2    
     }
   }  
  aopt = optimize(gcv,c(0,1),lam=lambda,mu=mu,delnrm2=delnrm2,maximum=FALSE,tol=1e-8) 
  alfmin = aopt$minimum
 }
alf = alfmin+1e-10
alffac = exp(-log(alf)/alfstp)   

#COMPUTE RECS ESTIMATE
done = 0
step = 1

while ((done == 0) & (step <= alfstp)) {

  #COMPUTE CMI MATRIX
  lamnew = lambda[1,] + alf
  sgvl = singval[1,]
  vvv1 = vvv[1,,]
  uuu1 = uuu[1,,]
  midmat = diag(sgvl/lamnew)
  cmimat = vvv1 %*% midmat %*% t(uuu1) %*% qhmat
  cmi[1,,] = cmimat
  if (phiscen==0) {
    for (i in 2:n) {
      cmi[i,,] = cmimat
     } 
   } else {
    for (i in 2:n) { 
      lamnew = lambda[i,] + alf
      sgvl = singval[i,]
      vvv1 = vvv[i,,]
      uuu1 = uuu[i,,]
      midmat = diag(sgvl/lamnew)
      cmimat = vvv1 %*% midmat %*% t(uuu1) %*% qhmat
      cmi[i,,] = cmimat
     }
   }
   
  #COMPUTE ESTIMATES 
  est0 = estval 
  estres = scoresoln(est0,cmi,psiw,cdwt,wwvec)
  estval = estres$b
  done = estres$igood
  
  # CHECK FOR EXTREME OUTLIERS
  maxabbet = max(abs(estval[2:r]))
  if (maxabbet > extrfac) {done=0}
  
  #INCREMENT STEP
  if (done==0) {
    step = step + 1
    alf = alffac*alf
   }

 } 

if (done == 1) {
  
#JACOBIAN MATRIX AND F MATRIX
dmat = scorejac(estval,cmi,psiw,cdwt,wwvec)
invd = solve(-dmat)
fmat = score(estval,cmi,psiw,cdwt,wwvec,2)  

#DERIVATIVE OF SCORE FUNCTION W.R.T. XI
if (ixiest == 1) {
for (j in 1:npxi) {

  #SET NEW XI
  xinew = xi
  xinew[j] = xinew[j] + didl

  #RECOMPUTE SETUPS
  settings = setup(xinew)
  psiw = settings$psiw
  phi = settings$phi
  cdwt = settings$cdwt
  wwvec = settings$wwvec
  uuu = settings$uuu
  vvv = settings$vvv
  singval = settings$singval
  lambda = settings$lambda
  
  #RECOMPUTE CMI MATRIX
  lamnew = lambda[1,] + alf
  sgvl = singval[1,]
  vvv1 = vvv[1,,]
  uuu1 = uuu[1,,]
  midmat = diag(sgvl/lamnew)
  cmimat = vvv1 %*% midmat %*% t(uuu1) %*% qhmat
  cmi[1,,] = cmimat
  if (phiscen==0) {
    for (i in 2:n) {
      cmi[i,,] = cmimat
     } 
   } else {
    for (i in 2:n) { 
      lamnew = lambda[i,] + alf
      sgvl = singval[i,]
      vvv1 = vvv[i,,]
      uuu1 = uuu[i,,]
      midmat = diag(sgvl/lamnew)
      cmimat = vvv1 %*% midmat %*% t(uuu1) %*% qhmat
      cmi[i,,] = cmimat
     }
   }

  #COMPUTE UBARD MATRIX 
  ubard[,j] = score(estval,cmi,psiw,cdwt,wwvec,1)/(n*didl)

 }
 }
 
#WRAP-UP
midlmat = fmat
if (ixiest == 1) {
  gmat = ubard%*%xicov%*%t(ubard)
  midlmat = midlmat + gmat
 }  
vbmat = invd%*%midlmat%*%invd
estbeta = trnsmat%*%estval
varbeta = trnsmat%*%vbmat%*%t(trnsmat)
vb = diag(varbeta)
estfin[2,] = estbeta
sefin[2,] = sqrt(vb)

}

results = list(estfin,sefin,naicov,varbeta,alfmin,step)
names(results) = c('estfin','sefin','naicov','varbeta','alfmin','step')
return(results)

} #END OF MAIN RECS ROUTINE

######################################################################################### 

#MAIN SECTION OF PROGRAM

# RESULTS FROM VALIDATION DATA

if (iscen == 1) {
  xwint = 1.22701
  xwslp = 0.56533
  npxi = 1
  ea = 0.81812/(xwslp^2)
  eb = NULL
 }
 
if (iscen == 2) {
  xwint = 1.22701
  xwslp = 0.56533
  npxi = 2
  ea =  0.58833/(xwslp^2)
  eb = 0.34970/(xwslp^2)
 }

xiest = c(ea,eb)    
xicov = matrix(0,npxi,npxi)

# MAIN DATA
data = read.table("main1.dat")
y = data[,1]
w = data[,2]
n = length(w)
meanw = mean(w)
w = (w-xwint)/xwslp
z = NULL
dim(y) = c(n,1)
dim(w) = c(n,1)
xi = xiest

#PRINT OUT INPUTS
inputs = cbind(n,nex,p,pz,kk1,kk2,dfr,ialmin,iscen,ixiest,npxi)
print(inputs)
print(cbind(xiest))

#SOME SETUPS
r = p + pz + 1
dfr1 = dfr-1
pdfr = dfr^p
pkk1 = kk1^p
pkk2 = kk2^p
phiscen = (iscen==3)
sfactr = qnorm(0.75,0,1)

#SET UP INDEX SELECTOR MATRICES

#digmatpdfr
dfixv = c(1:pdfr)
dim(dfixv) = c(pdfr,1)
digmatpdfr = digi(dfixv,dfr,p)

#digmatpkk1
pkixv1 = c(1:pkk1)
dim(pkixv1) = c(pkk1,1)
digmatpkk1 = digi(pkixv1,kk1,p)

#digmatpkk2
pkixv2 = c(1:pkk2)
dim(pkixv2) = c(pkk2,1)
digmatpkk2 = digi(pkixv2,kk2,p)

#SET UP MATRICES AND ARRAYS
wshift = rep(0,p)
wscal = rep(0,p)
uuu = array(rep(0,n*pkk2*pdfr),dim=c(n,pkk2,pdfr))
vvv = array(rep(0,n*pdfr*pdfr),dim=c(n,pdfr,pdfr))
singval = matrix(0,n,pdfr)
lambda = matrix(0,n,pdfr)

#NODES AND WEIGHTS FOR A-OPERATOR INTEGRAL
if (iscen <= 3) {
  gh1 = gauss.quad(kk1,kind="hermite")
  nodes1 = sqrt(2)*gh1$nodes
  qv1 = gh1$weights / sqrt(pi)
 }
if (iscen == 4) {
  kkcur = kk1/2
  gh1 = gauss.quad(kkcur,kind="laguerre")
  nod1a = gh1$nodes
  wt1a = gh1$weights
  nod1b = -nod1a[c(kkcur:1)]
  nodes1 = c(nod1b,nod1a) / sqrt(2)
  wt1b = wt1a[c(kkcur:1)]
  qv1 = c(wt1b,wt1a)/2
 }   
if (iscen > 4) {
  mccut = sqrt(5)
  qv1fac = 2/(sqrt(pi)*pchisq(5,1))
  kkcur = 2*kk1-1
  inodflg = 0
  while (inodflg == 0) {
    kkcur = kkcur+1
    ghcur = gauss.quad(kkcur,kind="hermite")
    nodcur = sqrt(2)*ghcur$nodes
    nodkeep = (nodcur >= 0) & (nodcur <= mccut)
    nodcnt = sum(nodkeep)
    if (nodcnt == (kk1+1)) {inodflg = 1}   
   }
  kkcur = kkcur - 1   
  ghcur = gauss.quad(kkcur,kind="hermite")
  nodcur = sqrt(2)*ghcur$nodes
  nodkeep = (nodcur >= 0) & (nodcur <= mccut)
  nodes1 = nodcur[nodkeep]
  wtcur = ghcur$weights
  qv1 = qv1fac*wtcur[nodkeep]
  nodes1 = ((nodes1^2)-mdchmean)/mdchsd  
 }

#NODES AND WEIGHTS FOR L2 NORM
gh2 = gauss.quad(kk2,kind="hermite")
nodes2 = sqrt(2)*gh2$nodes
qv2 = gh2$weights / sqrt(pi)
qv2new = rep(1,pkk2)
for (j in 1:pkk2) {
for (ip in 1:p) {
  ix = digmatpkk2[j,ip]
  qv2new[j] = qv2new[j]*qv2[ix]
 } 
 }
sqqv2new = sqrt(qv2new)
qhmat = diag(sqqv2new)  

#SET UP ORTHOGONAL POLYNOMIAL BASIS
coefmat = pbas(dfr)

# BEGIN ESTIMATION PROCESS
 
#RESCALE W TO MEAN 0 AND VARIANCE 1
for (l in 1:p) {
  wcur = w[,l]
  wcen = mean(wcur)
  wscf = sd(wcur)
  wscal[l] = 1/wscf
  wshift[l] = wcen/wscf
  wcur = wscal[l]*wcur - wshift[l]
  dim(wcur) = c(n,1)
  w[,l] = wcur
 }
dtrm = c(1,wscal,rep(1,pz)) 
trnsmat = diag(dtrm)
trnsm1 = -wshift
trnsm1 = c(1,trnsm1,rep(0,pz))
trnsmat[1,] = trnsm1
dim(wscal) = c(p,1)
   
#RUN RECS
results = recs(y,w,z)
print(results)
estfin = results$estfin
naicov = results$naicov
varbeta = results$varbeta
print("naive std errors")
print(sqrt(diag(naicov)))
print("recs std errors")
print(sqrt(diag(varbeta)))

### THE END ###
